#!/usr/bin/env python3
"""
Unproprietary
Gzip, PNG, Zlib, BZip2, XZ, RAR, Zip, TAR, ELF, SWF, Ogg, MP3, WAV, PDF, BitTorrent, AVI, LZMA, FLAC, Unicode, Nullsoft, Hammerwatch, BMP, TGA finder
by Javantea
Mar 24, 2012

Similar to Hachior but less buggy.
Similar to file but doesn't give up on the header.
SWF files use zlib, 

TODO:
NetPBM
MPEG4
MIDI
PGP
PostScript
EPub (we already do zip, but find the right header)
FLAC

Unknown file size prevents us from properly extracting many file types.

TODO:
improve upon zlib false positives a bit more.
improve upon mp3 false positives.
improve PDF find without header.

Note that PDF files do not require a header and have no recognizable required 
format. Certain elements are found but are silly.
"""
from __future__ import print_function
import struct
import zlib
import mmap
import sys
import binascii
from zlibPy import outputZlib
try:
	import collections
except ImportError:
	# If they don't have collections, they're running very old Python.
	print("Upgrade Python if you want better quality results.")
	collections = None
#end try

def usage():
	print("Usage: find_compress.py [-o offset] [-m] [-v] [-l length] file.ext [...] > file.dat")
	print("    -o offset Skip to file offset, default 0")
	print("    -m	Multiple solutions")
	print("    -v	Verbose")
	print("    -l length Read length, default 1024 kB")
	print("    -d dir    Output directory")
#end def usage()

class POSSIBLE:
	matchValue = 23
class UNLIKELY:
	matchValue = 13

def findGzip(f, fpos):
	"""
	Given N bytes of data, determine whether it's a GZip.
	GZip format is:
	magic 0x1f8b
	CM    8
	FLG   0 or 8 usually
	timestamp
	"""
	data = f[fpos:fpos+4]
	if data == None or len(data) < 2: return False
	if data[0:1] != b'\x1f' or data[1:2] != b'\x8b': return False
	if len(data) < 3: return True
	if data[2:3] != b'\x08': return POSSIBLE
	if len(data) < 4: return True
	if data[3:4] != b'\x00' and data[3:4] != b'\x08': return POSSIBLE
	return True
#end def findGzip(f, fpos)

def parseGzip(f, fpos):
	"""
	magic 0x1f8b
	CM    8
	FLG   0 or 8 usually
	MTIME timestamp
	XFL   2 or 4
	OS    0, 11, or 3
	if FLG & 4 == 4:
		XLEN 16-bit
		Extra field
	if FLG & 8 == 8:
		Original filename zero terminated
	if FLG & 16 == 16:
		file comment zero terminated
	if FLG & 2 == 2:
		CRC16 for the gzip header
	compressed data
	CRC32
	"""
	whichDict = dict
	if collections and 'OrderedDict' in collections.__all__:
		whichDict = collections.OrderedDict
	#end if
	output = whichDict()
	output['magic'] = f[fpos:fpos+2]
	output['CM'] = f[fpos+2]
	FLG = f[fpos+3]
	output['FLG'] = FLG
	pos = fpos+3
	if FLG & 4 == 4:
		XLEN = struct.unpack('>H', output[pos:pos+2])[0]
		output['XLEN'] = XLEN
		output['extra'] = f[pos+2:pos+2+XLEN]
		pos += XLEN + 2
	if FLG & 8 == 8:
		endfn = f.find(b'\x00', pos)
		if endfn == -1:
			# We can't find the end, that's an error.
			return output
		#end if
		output['filename'] = f[pos:endfn]
		pos = endfn + 1
	if FLG & 16 == 16:
		endfn = f.find(b'\x00', pos)
		if endfn == -1:
			# We can't find the end, that's an error.
			return output
		#end if
		output['comment'] = f[pos:endfn]
		pos = endfn + 1
	if FLG & 2 == 2:
		CRC16 = struct.unpack('>H', output[pos:pos+2])[0]
		output['CRC16'] = CRC16
		pos += 2
	
	if len(f) < 8+pos: return output
	mtime = struct.unpack('>I', f[pos+4:pos+8])[0]
	output['MTIME'] = mtime
	pos += 4
	# TODO: get the rest.
	output['compressed_begin'] = f[pos:pos+20]
	return output
#end def parseGzip(f, fpos)

def findZlib(f, fpos):
	"""
	Given N bytes of data, determine whether it's a Zlib.
	Zlib format is:
	CINFO 7 (base-2 logarithm of LZ77 window size)
	CM    8 (deflate)
	
	The FCHECK value must be such that CMF and FLG, when viewed as
	a 16-bit unsigned integer stored in MSB order (CMF*256 + FLG),
	is a multiple of 31.
	
	Stupid mistake 'x ' is ASCII and a common occurrence. Therefore we must 
	say that 'x ' is possible but unlikely...
	"""
	data = f[fpos:fpos+10]
	if data == None or len(data) < 2: return False
	# Pretty much all zlib is 'x', anomalies include 'h' and so forth.
	# 
	if (data[0]) & 0xf != 8: return False
	CMF = (data[0])
	FLG = (data[1])
	# Easy FCHECK
	mod31_fcheck = ((CMF<<8) + FLG) % 31
	if mod31_fcheck != 0: return False

	CM = (CMF & 0xf)
	# Already checked CM above
	CINFO = (CMF & 0xf0) >> 4
	# This is a limitation of zlib, https://www.ietf.org/rfc/rfc1950.txt
	if CINFO > 7:
		return False
	#end if
	# TODO: BITS, NEEDBITS, and DROPBITS would actually be useful here =)
	pos = 2
	fdict = (FLG & 0x20) >> 5
	dictid = None
	if fdict == 1:
		# TODO: Use DICTID maybe?
		dictid = data[2:6]
		pos += 4
	#end if
	# Worthless data. 0 = fastest, 1 = fast, 2 = default, 3 = slowest
	flevel = (FLG & 0xc0) >> 6
	
	if len(data) <= pos: return True

	# DEFLATE
	# BTYPE reserved (error) https://www.ietf.org/rfc/rfc1951.txt pp. 9
	btype = (data[pos] & 0x7) >> 1
	if btype == 0x3: return False

	# We could continue parsing the whole thing, but let's avoid that =)
	# See zlibPy.outputZlib for the full implementation of zlib in Python.
	
	return True
#end def findZlib(f, fpos)

def findBZip2(f, fpos):
	"""
	Given N bytes of data, determine whether it's a BZip2.
	BZip2 format is:
	magic BZh
	# byte  3 very commonly 9 but can be:
	1 1
	2 1
	3 1
	5 225
	6 2
	9 1697
	# byte  4-9 1AY&SY
	"""
	data = f[fpos:fpos+10]
	if data == None or len(data) < 3: return False
	if data[:3] != b'BZh': return False
	if data[4:10] == b'1AY&SY': return True
	# If there isn't this obvious signature we don't have a 100% guarantee.
	return POSSIBLE
#end def findBZip2(f, fpos)

def findXZ(f, fpos):
	"""
	Given N bytes of data, determine whether it's a XZ.
	XZ format is:
	magic 0xfd 7zXZ 0x00
	hist_xz1.txt is incredible.
	"""
	data = f[fpos:fpos+6]
	if data == None or len(data) < 4: return False
	# Extra code in case we find a shrunken one
	if len(data) < 6:
		if data[:4] != b'\xfd7zX': return False
		return POSSIBLE
	#end if
	if data[:6] != b'\xfd7zXZ\x00': return False
	return True
#end def findXZ(f, fpos)

def findPNG(f, fpos):
	"""
	Given N bytes of data, determine whether it's a PNG.
	PNG format is:
	magic \x89PNG
	0d 0a 1a 0a
	00 00 00 0d
	IHDR
	"""
	data = f[fpos:fpos+8]
	if data == None or len(data) < 4:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:4] != b'\x89PNG': return False
	if len(data) < 8:
		return POSSIBLE
	#end if
	if data[4:8] == b'\x0d\x0a\x1a\x0a': return True	
	return False
#end def findPNG(f, fpos)

def parsePNG(f, fpos):
	"""
	Given N bytes of data that passes findPNG, find the end and parts of a 
	PNG.
	PNG format is:
	magic \x89PNG
	0d 0a 1a 0a
	chunk length 00 00 00 0d
	IHDR
	Width 	4 bytes
	Height 	4 bytes
	Bit depth 	1 byte
	Colour type 	1 byte
	Compression method 	1 byte
	Filter method 	1 byte
	Interlace method 	1 byte
	"""
	flen = len(f) - fpos
	whichDict = dict
	if collections and 'OrderedDict' in collections.__all__:
		whichDict = collections.OrderedDict
	#end if
	out = whichDict()
	out['sig'] = f[fpos:fpos+4]
	if flen < 8:
		return out
	#end if
	out['sig2'] = f[fpos+4:fpos+8]
	if flen < 12:
		return out
	#end if
	IHDR_len, = struct.unpack('>I', f[fpos+8:fpos+12])
	out['IHDR_len'] = IHDR_len
	if flen < 16:
		return out
	#end if
	out['IHDR'] = f[fpos+12:fpos+16]
	if flen < 29:
		return out
	#end if
	width, = struct.unpack('>I', f[fpos+16:fpos+20])
	out['width']  = width
	height, = struct.unpack('>I', f[fpos+20:fpos+24])
	out['height'] = height
	bit_depth = (f[fpos+24])
	out['bit_depth'] = bit_depth
	color_type = (f[fpos+25])
	out['color_type'] = color_type
	compression_method = (f[fpos+26])
	out['compression_method'] = compression_method
	filter_method = (f[fpos+27])
	out['filter_method'] = filter_method
	interlace_method = (f[fpos+28])
	out['interlace_method'] = interlace_method
	if flen < 33:
		return out
	#end if
	IHDR_crc32, = struct.unpack('>i', f[fpos+29:fpos+33])
	out['IHDR_crc32'] = IHDR_crc32
	IHDR_cmp_crc32 = zlib.crc32(f[fpos+12:fpos+29])
	out['IHDR_cmp_crc32'] = IHDR_cmp_crc32
	if IHDR_crc32 != IHDR_cmp_crc32:
		out['Error'] = ['IHDR crc32 does not match']
	#end if
	out['chunks'] = []
	cpos = fpos+33
	ctype = 'IHDR'
	while ctype != 'IEND':
		if flen < cpos+4: break
		clen, = struct.unpack('>I', f[cpos:cpos+4])
		if flen < cpos+8+clen+4: break
		ctype = f[cpos+4:cpos+8]
		cdata = f[cpos+8:cpos+8+clen]
		c_crc32, = struct.unpack('>i', f[cpos+8+clen:cpos+8+clen+4])
		actual_crc32 = zlib.crc32(f[cpos+4:cpos+8+clen])
		out['chunks'].append((clen, ctype, cdata, c_crc32, actual_crc32))
		cpos += 8 + clen + 4
	#loop
	out['end'] = cpos
	return out
#end def parsePNG(f, fpos)

def outputPNG(f, fpos):
	"""
	Given N bytes of data that passes findPNG, find the end and parts of a 
	PNG.
	PNG format is:
	magic \x89PNG
	0d 0a 1a 0a
	chunk length 00 00 00 0d
	IHDR
	Width 	4 bytes
	Height 	4 bytes
	Bit depth 	1 byte
	Colour type 	1 byte
	Compression method 	1 byte
	Filter method 	1 byte
	Interlace method 	1 byte
	"""
	# TODO: test me!
	flen = len(f)
	cpos = fpos + 33
	ctype = 'IHDR'
	while ctype != 'IEND':
		if flen < cpos+4: break
		clen, = struct.unpack('>I', f[cpos:cpos+4])
		if flen < cpos+8+clen+4: break
		ctype = f[cpos+4:cpos+8]
		#cdata = f[cpos+8:cpos+8+clen]
		#c_crc32, = struct.unpack('>i', f[cpos+8+clen:cpos+8+clen+4])
		#actual_crc32 = zlib.crc32(f[cpos+4:cpos+8+clen])
		cpos += 8 + clen + 4
	#loop
	end = cpos - fpos
	return f[fpos:fpos+end]
#end def outputPNG(f, fpos)

def findZIP(f, fpos):
	"""
	Given N bytes of data, determine whether it's a ZIP.
	ZIP format is:
	magic PK 0x3 0x4
	version needed to extract       2 bytes
	general purpose bit flag	2 bytes
	compression method	      2 bytes
	last mod file time	      2 bytes
	last mod file date	      2 bytes
	crc-32			  4 bytes
	compressed size		 4 bytes
	uncompressed size	       4 bytes
	file name length		2 bytes
	extra field length	      2 bytes

	A funny exception about ZIP is that it can be at any offset and still 
	decompress.
	"""
	data = f[fpos:fpos+4]
	if data == None or len(data) < 4:
		#print 'Error: somehow data is not long enough:', len(data)
		return {}
	#end if
	if data[:4] == b'PK\x03\x04': return True
	return False
#end def findZIP(f, fpos)

def parseZIP(f, fpos):
	"""
	Given N bytes of data that passes findZIP, find the end and parts of a 
	ZIP.
	ZIP format is:
	magic PK 0x3 0x4
	version needed to extract       2 bytes
	general purpose bit flag	2 bytes
	compression method	      2 bytes
	last mod file time	      2 bytes
	last mod file date	      2 bytes
	crc-32			  4 bytes
	compressed size		 4 bytes
	uncompressed size	       4 bytes
	file name length		2 bytes
	extra field length	      2 bytes

	A funny exception about ZIP is that it can be at any offset and still 
	decompress.
	"""
	flen = len(f)
	if flen < 4:
		return False
	#end if
	whichDict = dict
	if collections and 'OrderedDict' in collections.__all__:
		whichDict = collections.OrderedDict
	#end if
	out = whichDict()
	out['sig'] = f[fpos:fpos+4]
	return out
#end def parseZIP(f, fpos)

def findGIF(f, fpos):
	"""
	Given N bytes of data, determine whether it's a GIF.
	GIF format is:
	magic GIF
	version 89a or 87a
	This will give false positives because it's ASCII and it's part of docs 
	describing GIF including this one.
	"""
	data = f[fpos:fpos+6]
	if data == None or len(data) < 4:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:3] != b'GIF': return False
	if len(data) < 6: return POSSIBLE
	if data[3:6] == b'89a' or data[3:6] == b'87a': return True
	return False
#end def findGIF(f, fpos)

def parseGIF(f, fpos):
	"""
	Given N bytes of data, find the end and parse the parts of the GIF.
	GIF format is:
	magic GIF
	version 89a or 87a
	Logical screen descriptor
		width little endian short
		height little endian short
		GCT byte
		background byte
		default pixel aspect ratio byte
		color[256] 3 bytes each
	Graphic Control Extension 2 bytes
		GCE len byte
		transparency byte
		animation delay little endian short (not used)
		which color is transparent byte
		end of GCE byte
	Image Descriptor
		NW Corner 2 little endian unsigned shorts
		Width and Height 2 little endian unsigned shorts
		local color table byte
	Start of image byte
	LZW length byte
	LZW data N bytes
	end of image data byte
	GIF file terminator byte
	
	This will give false positives because it's ASCII and it's part of docs 
	describing GIF including this one.
	"""
	whichDict = dict
	# Ordered dict is way better for this purpose. TODO: improve str function.
	if collections and 'OrderedDict' in collections.__all__:
		whichDict = collections.OrderedDict
	#end if
	data = whichDict()
	data['magic'] = f[fpos:fpos+3]
	flen = len(f)
	if flen < 6: return data
	version = f[fpos+3:fpos+6]
	data['version'] = version
	if version != b'89a' and version != b'87a': return data
	if flen < 13: return data
	lsd_width, lsd_height = struct.unpack('<HH', f[fpos+6:fpos+10])
	data['lsd_width'] = lsd_width
	data['lsd_height'] = lsd_height
	data['GCT'] = f[fpos+10:fpos+11]
	data['background'] = f[fpos+11:fpos+12]
	data['aspect_ratio'] = f[fpos+12:fpos+13]
	data['colors'] = [f[fpos+13+(i*3):fpos+13+(i*3)+3] for i in range(256)]
	
	return data
#end def findGIF(f, fpos)

def findBMP(f, fpos):
	"""
	Given N bytes of data, determine whether it's a BMP.
	BMP format is:
	magic BM | BA | CI | CP | IC | PT
	size little endian dword
	reserved little endian short
	reserved little endian short
	offset little endian dword
	False positives are mitigated well by validating the PC bitmap, Windows 3.x format.
	"""
	data = f[fpos:fpos+2]
	if data == None or len(data) < 2:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:2] in [b'BM', b'BA', b'CI', b'CP', b'IC', b'PT']:
		if f[fpos+14:fpos+16] == b'\x28\x00':
			return True
		#end if
	#end if
	return False
#end def findBMP(f, fpos)

def outputBMP(f, fpos):
	"""
	Given N bytes of data that passes findBMP, find the end of BMP. This will 
	be very bad for false positives.
	"""
	# TODO: test me!
	flen = len(f) - fpos
	end = struct.unpack('<I', f[2:6])[0]
	return f[fpos:fpos+end]
#end def outputBMP(f, fpos)


def findTGA(f, fpos):
	"""
	Given N bytes of data, determine whether it's a BMP.
	Targa (TGA) format is:
	sig dword
	sig & 0xfff7ffff == 0x01010000
	sig & 0xfff7ffff == 0x00020000
	sig & 0xfff7ffff == 0x00030000
	According to file-5.22/magic/Magdir/images
	This will give false positives because it's badly designed mostly null bits magic.
	"""
	data = f[fpos:fpos+5]
	if data == None or len(data) < 5:
		return False
	#end if
	sig = struct.unpack('>I', data[1:5])[0]
	if sig in [0x01010000, 0x00020000, 0x00030000]: return True
	return False
#end def findTGA(f, fpos)


def findMP3(f, fpos):
	"""
	Given N bytes of data, determine whether it's an MP3.
	MP3 format is:
	magic ID3
	unknown variable \x03 also possible \x02 \x04 or others.
	\x00\x00\x00
	\x00 or non-zero
	2 unknown bytes with slightly higher likelihood of \x12\x64
	
	magic 
	
	FIXME: False positive in CAM_Installer V1.0.0.exe and many others.
	These false positives are actually part of the MP3 spec. Since the header 
	is so commonly found in all random (compressed) files, it practically 
	guarantees false positives.
	"""
	data = f[fpos:fpos+7]
	if data == None or len(data) < 3:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:3] == b'ID3': 
		if len(data) < 7: return POSSIBLE
		if data[3:7] == b'\x03\x00\x00\x00' or data[3:7] == b'\x02\x00\x00\x00' or data[3:7] == b'\x04\x00\x00\x00': return True
		return False
	#end if
	hm = struct.unpack('>H', data[:2])[0]
	# These false positive checks learned from libmpg123/parse.c head_check()
	# HDR_SYNC 0xffe00000
	if (hm & 0xFFE0) != 0xFFE0: return False
	# MPEG Audio version id == reserved
	if (hm & 0x18) == 0x8: return False
	# HDR_LAYER 0x00060000
	if (hm & 0x6) == 0: return False
	hm2 = struct.unpack('>H', data[2:4])[0]
	# HDR_BITRATE 0x0000f000
	if (hm2 & 0xf000) == 0xf000: return False
	# HDR_SAMPLERATE 0x00000c00
	if (hm2 & 0xc00) == 0xc00: return False
	# Emphasis == reserved
	if (hm2 & 0x3) == 0x2: return False
	# FIXME: This is way too common an occurrence
	# FIXME: it rules out a bunch of legitimate modes.
	if (hm & 0xFFFE) == 0xFFFA:
		return True
	#end if
	if (hm & 0xFFFE) == 0xFFF2:
		return True
	#end if
	if (hm & 0xFFFE) == 0xFFE2:
		return True
	#end if
	return False
#end def findMP3(f, fpos)

def findWAV(f, fpos):
	"""
	Given N bytes of data, determine whether it's an WAVE.
	WAVE format is:
	magic RIFF
	length int32_t
	magic WAVE
	"""
	data = f[fpos:fpos+12]
	if data == None or len(data) < 3:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:4] != b'RIFF': return False
	if len(data) < 12: return POSSIBLE
	if data[8:12] == b'WAVE': return True
	# RIFFs that aren't WAVE format may still be a wav
	return POSSIBLE
#end def findWAV(f, fpos)

def parseWAV(f, fpos):
	"""
	Given N bytes of data that passes findWAV, find the end and parts of a 
	WAV.
	magic RIFF
	length int32_t
	magic WAVE
	"""
	if f == None or (len(f) < (fpos+16)):
		return None
	#end if
	whichDict = dict
	if collections and 'OrderedDict' in collections.__all__:
		whichDict = collections.OrderedDict
	#end if
	out = whichDict()
	length, = struct.unpack('<I', f[fpos+4:fpos+8])
	out['RIFF'] = f[fpos:fpos+4]
	out['format'] = f[fpos+8:fpos+12]
	out['end'] = length
	out['Subchunk1ID'] = f[fpos+12:fpos+16]
	if (len(f) < (fpos+44)):
		return out
	#end if
	Subchunk1Size, AudioFormat, NumChannels, SampleRate, ByteRate, BlockAlign, BitsPerSample = struct.unpack('<IHHIIHH', f[fpos+16:fpos+36])
	out.update({'Subchunk1Size':Subchunk1Size, 'AudioFormat':AudioFormat, 'NumChannels':NumChannels, 'SampleRate':SampleRate, 'ByteRate':ByteRate, 'BlockAlign':BlockAlign, 'BitsPerSample':BitsPerSample})
	out['Subchunk2ID'] = f[fpos+36:fpos+40]
	Subchunk2Size, = struct.unpack('<I', f[fpos+40:fpos+44])
	out['Subchunk1Size'] = Subchunk2Size
	
	if (length + fpos) > len(f):
		print('Warning: length is longer than data len', fpos + length, '>', len(f))
	#end if
	return out
#end def parseWAV(f, fpos)

def outputWAV(f, fpos):
	"""
	Given N bytes of data that passes findWAV, find the end and parts of a 
	WAV.
	magic RIFF
	length int32_t
	magic WAVE
	"""
	length = struct.unpack('<I', f[fpos+4:fpos+8])[0] + 4
	# TODO: reinsert
	#if length > len(data):
	#	print 'Warning: length is longer than data len', length, '>', len(data)
	#	return data
	#end if
	return f[fpos:fpos+length]
#end def outputWAV(f, fpos)

def findAVI(f, fpos):
	"""
	Given N bytes of data, determine whether it's an AVI.
	AVI format is:
	magic RIFF
	length int32_t
	magic AVI 
	"""
	data = f[fpos:fpos+12]
	if data == None or len(data) < 3:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:4] != b'RIFF': return False
	if len(data) < 12: return POSSIBLE
	if data[8:12] == b'AVI ': return True
	# RIFFs that aren't AVI format may still be a video
	return POSSIBLE
#end def findAVI(f, fpos)


def findOgg(f, fpos):
	"""
	Given N bytes of data, determine whether it's an Ogg.
	Ogg format is:
	magic Oggs
	version \x00
	header_type_flag \x02 (first)
	absolute granule position uint64_t
	stream_serial_number uint32_t
	page_sequence_no uint32_t
	page_checksum uint32_t
	page_segments 
	segment_table
	data
	"""
	data = f[fpos:fpos+14]
	if data == None or len(data) < 4:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:4] != b'OggS': return False
	if len(data) < 14: return POSSIBLE
	if data[4:14] == b'\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00': return True
	if data[4:14] == b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00': return True
	if data[4:5] == b'\x00': return POSSIBLE
	#print 'hi', repr(data[4:14])
	return False
#end def findOgg(f, fpos)

def parseOGG(f, fpos):
	"""
	Given N bytes of data that passes findPNG, find the end and parts of a 
	OGG.
	OGG format is:
	magic Oggs
	version \x00
	header_type_flag \x02 (first)
	absolute granule position uint64_t
	stream_serial_number uint32_t
	page_sequence_no uint32_t
	page_checksum uint32_t
	page_segments 
	segment_table
	data
	"""
	if f == None or (len(f) < (fpos + 4)):
		#print 'Error: somehow data is not long enough:', len(data)
		return {}
	#end if
	whichDict = dict
	if collections and 'OrderedDict' in collections.__all__:
		whichDict = collections.OrderedDict
	#end if
	out = whichDict()
	out['sig'] = f[:4]
	if len(f) < (fpos + 5):
		return out
	#end if
	out['pages'] = []
	pos = fpos
	while pos + 26 < len(f):
		page = {}
		page['capture_pattern'] = f[pos:pos+4]
		if page['capture_pattern'] != b'OggS':
			print('Error: Bad OggS capture pattern', repr(page['capture_pattern']))
			break
		#end if
		page['stream_structure_version'] = f[pos+4]
		page['header_type_flag'] = f[pos+5]
		page['fresh'] = page['header_type_flag'] & 0x1
		page['first'] = page['header_type_flag'] & 0x2
		page['last'] = page['header_type_flag'] & 0x4
		# Anything except 1,2,4 are an error case
		if page['header_type_flag'] & 0xf8: page['error'] = True
		page['absolute_granule_position'], = struct.unpack('>Q', f[pos+6:pos+14])
		page['stream_serial_number'], = struct.unpack('>I', f[pos+14:pos+18])
		page['page_sequence_no'], = struct.unpack('>I', f[pos+18:pos+22])
		page['page_checksum'], = struct.unpack('>I', f[pos+22:pos+26])
		page['page_segments'] = f[pos+26]
		page_segs = []
		data_size = 0
		for i in range(page['page_segments']):
			v = f[pos+27+i]
			page_segs.append(v)
			data_size += v
		#next i
		page['segment_table'] = page_segs
		page['data_size'] = data_size
		out['pages'].append(page)
		pos += 27 + page['page_segments'] + data_size
		if page['last']: break
	#loop
	out['end'] = pos
	return out
#end def parseOGG(f, fpos)


def outputOGG(f, fpos):
	"""
	Given N bytes of data that passes findOGG, find the end and parts of a 
	OGG.
	OGG format is:
	magic Oggs
	version \x00
	header_type_flag \x02 (first)
	absolute granule position uint64_t
	stream_serial_number uint32_t
	page_sequence_no uint32_t
	page_checksum uint32_t
	page_segments 
	segment_table
	data
	FIXME: need to duplicate the results of split_isb.py
	python find_compress1.py -d samusic/ -m ~/psychonauts/WorkResource/Sounds/samusic.isb 
	FIXME: not working yet.
	"""
	#if data == None or len(data) < 4:
	#	#print 'Error: somehow data is not long enough:', len(data)
	#	return ''
	#end if
	# TODO: test me!
	pos = 0
	# FIXME: need to know file_size
	#while pos + 26 < len(data):
	while True:
		header_type_flag = f[fpos+pos+5]
		fresh = header_type_flag & 0x1
		first = header_type_flag & 0x2
		last = header_type_flag & 0x4
		page_segments = f[fpos+pos+26]
		page_segs = []
		data_size = 0
		for i in range(page_segments):
			v = f[fpos+pos+27+i]
			page_segs.append(v)
			data_size += v
		#next i
		pos += 27 + page_segments + data_size
		# We might want to turn this off but...
		if last: break
	#loop
	end = pos
	return f[fpos:fpos+end]
#end def outputOGG(f, fpos)

def findFLAC(f, fpos):
	"""
	Given N bytes of data, determine whether it's an FLAC.
	FLAC format is:
	magic fLaC
	STREAMINFO
		header
			<1> last-metadata-block
			<7> BLOCK_TYPE (STREAMINFO=0)
			<24> Length of metadata to follow
	...
	This could be done with 2 if statements instead if we wanted to.
	"""
	data = f[fpos:fpos+8]
	if data == None or len(data) < 4:
		return False
	#end if
	if data[:4] != b'fLaC': return False
	if len(data) < 8: return POSSIBLE
	x = struct.unpack('>I', data[4:8])[0]
	BLOCK_TYPE = (x >> 24) & 0x7f
	# A very harsh and unforgiving test past the fourCC requirement.
	if BLOCK_TYPE != 0: return POSSIBLE
	Last_metadata_block = (x >> 31)
	METADATA_SIZE = x & 0xffffff
	if METADATA_SIZE != 34: return POSSIBLE
	
	# It's very likely to be FLAC.
	return True
#end def findFLAC(f, fpos)

def findTAR(f, fpos):
	"""
	Given N bytes of data, determine whether it's a TAR.
	TAR format is:
	filename[100]
	mode[8]
	uid[8]
	gid[8]
	size[12]
	mtime[12]
	check_sum[8]
	type[1]
	lname[100]
	magic[8]
	To get the magic, you need 257+8 bytes a ridiculously large amount
	"""
	data = f[fpos:fpos+265]
	if data == None or len(data) < 4:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if len(data) >= 265:
		return (data[257:265] == b'ustar  \x00')
	#end if
	return False
	"""# This will create a lot of false positives really need the ustar.
	# Also false negatives with unicode filenames, gah
	eofn = False
	for i in range(min(100, len(data))):
		if eofn:
			if data[i] != '\x00': return False
			continue
		#end if
		c = ord(data[i])
		if (c < 32 or c > 126):
			if c != 0: return False
			eofn = True
			#print 'filename', repr(data[:i])
		#end if
	#next i
	if len(data) >= 101:
		if data[100] != '0': return False
	#end if
	return True"""
#end def findTAR(f, fpos)

def findJPEG(f, fpos):
	"""
	Given N bytes of data, determine whether it's a JPEG.
	JPEG format is:
	SOI ff d8
	APP ff e0-ef means APP 0-15
	DHQ ff db
	and so forth, totally mixed up and totally readable.
	the only thing we can count on are ff02-ffbf being reserved for now.
	"""
	data = f[fpos:fpos+9]
	if data == None or len(data) < 3:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:3] != b'\xff\xd8\xff': return False
	# APP1 == e1 is readable.
	app = (data[3])
	if app < 0xc0: return False
	if len(data) < 9: return POSSIBLE
	if data[6:11] == b'JFIF\x00': return True
	# SEE http://www.faqs.org/faqs/jpeg-faq/part1/section-15.html
	# "you probably have a not-quite-JFIF JPEG file"
	# FIXME: This is too little to go on see above, but some files are 
	# seriously messed up.
	return True
#end def findJPEG(f, fpos)

def parseJPEG(f, fpos):
	"""
	Given N bytes of data that passes findJPEG, find the end and parts of a 
	JPEG.
	http://www.w3.org/Graphics/JPEG/
	JPEG format is:
	SOI ff d8
	APP ff e0-ef means APP 0-15
	DHQ ff db
	EOI ff d9
	and so forth, totally mixed up and totally readable.
	the only thing we can count on are ff02-ffbf being reserved for now.
	"""
	markers = {b'\xff\xd8':'SOI', b'\xff\xc0':'SOF0', b'\xff\xc0':'SOF0', 
	'\xff\xc0':'SOF1', b'\xff\xc2':'SOF2', b'\xff\xc3':'SOF3', 
	'\xff\xc5':'SOF5', b'\xff\xc6':'SOF6', 	'\xff\xc7':'SOF7', 
	'\xff\xc8':'JPG', b'\xff\xc9':'SOF9', b'\xff\xca':'SOF10', 
	'\xff\xcb':'SOF11', b'\xff\xcd':'SOF13', b'\xff\xce':'SOF14', 
	'\xff\xcf':'SOF15', b'\xff\xc4':'DHT', b'\xff\xcc':'DAC', 
	'\xff\xd0':'RST0', b'\xff\xd1':'RST1', b'\xff\xd2':'RST2', 
	'\xff\xd3':'RST3', b'\xff\xd4':'RST4', b'\xff\xd5':'RST5', 
	'\xff\xd6':'RST6',  '\xff\xd7':'RST7', b'\xff\xda':'SOS', 
	'\xff\xdb':'DQT', b'\xff\xdc':'DNL', b'\xff\xdd':'DRI', 
	'\xff\xde':'DHP', b'\xff\xdf':'EXP', b'\xff\xe0':'APP0', 
	'\xff\xfe':'COM', b'\xff\x00':'ff', b'\xff\xd9':'EOI'}
	for i in range(1, 16):
		markers[b'\xff' + bytes([0xe0+i])] = 'APP%i' % i
	#next i
	for i in range(14):
		markers[b'\xff' + bytes([0xf0+i])] = 'JPG%i' % i
	#next i
	if f == None or (len(f) < (fpos + 3)):
		#print 'Error: somehow data is not long enough:', len(data)
		return {}
	#end if
	whichDict = dict
	if collections and 'OrderedDict' in collections.__all__:
		whichDict = collections.OrderedDict
	#end if
	out = whichDict()
	out['soi'] = f[fpos:fpos+2]
	out['commands'] = []
	#blocks = data.split('\xff')
	# Starts with ff, so pop.
	#if blocks[0] == '': blocks.pop(0)
	# Remove and \xff\x00
	j = 0
	pos = fpos+2
	#for block in blocks:
	while pos+2 <= len(f):
		command = f[pos:pos+2]
		if pos+4 < len(f):
			readLen, = struct.unpack('>H', f[pos+2:pos+4])
		else:
			readLen = 0
		#end if
		
		if command in markers:
			marker = markers[command]
		else:
			marker = 'invalid'
		#end if
		actualLen = readLen
		if marker == 'SOS':
			# SOS has length of scan header instead of length of data.
			# Length of scan header
			Ls = readLen
			# Number of image components in scan
			Ns = f[pos+4]
			# Scan for ff non zero
			x = pos + 3
			foundX = None
			while x+2 <= len(f):
				if f[x] == 255 and f[x+1] != 0:
					#print 'found marker', repr(data[x:x+2])
					foundX = x
					break
				#end if
				x += 1
			#loop
			if foundX == None:
				actualLen = x - pos - 2
				print('Warning: failed to find end of scan')
			else:
				actualLen = foundX - pos - 2
			#end if
		#end if
		out['commands'].append({'cmd':binascii.hexlify(command),
			'marker':marker, 
			'len':readLen, 
			'data':f[pos+2:pos+2+actualLen]})
		if command == b'\xff\xd9': break
		pos += 2+actualLen
	#next block
	# FIXME: Alas this doesn't work unless you have SOS.
	if b'\xff\xd9' in f:
		out['end'] = f.index(b'\xff\xd9') + 2
	#end if
	return out
#end def parseJPEG(f, fpos)

def outputJPEG(f, fpos):
	"""
	Given N bytes of data that passes findJPEG, find the end and parts of a 
	JPEG.
	SOI ff d8
	EOI ff d9
	"""
	end_found = f.find(b'\xff\xd9', fpos)
	if end_found >= 0:
		return f[fpos:fpos + end_found + 2]
	#end if
	# TODO: note that we did not find EOI
	return f[fpos:fpos+10000]
#end def outputJPEG(f, fpos)

def findSWF(f, fpos):
	"""
	Given N bytes of data, determine whether it's a SWF.
	SWF format is:
	magic CWS or FWS
	version between 1 and 11
	file_size < MAX_FILE_SIZE = 50 * 1024 * 1024
	zlib
	"""
	data = f[fpos:fpos+4]
	if data == None or len(data) < 4:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:3] != b'CWS' and data[:3] != b'FWS': return False
	version = (data[3])
	if version <= 0 or version > 11: return False
	return True
#end def findSWF(f, fpos)

def findRAR(f, fpos):
	"""
	Given N bytes of data, determine whether it's a RAR.
	RAR format is:
	signature 0x52 0x61 0x72 0x21 0x1A 0x07 0x00
	'Rar!\x1a\x07\x00'
	There are more file headers now.
	"""
	data = f[fpos:fpos+7]
	if data == None or len(data) < 4:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:4] != b'Rar!': return False
	# Rar! is not a sure thing because it's ASCII and it's part of docs 
	# describing rar including this one.
	if len(data) < 7: return POSSIBLE
	if data[4:7] == b'\x1a\x07\x00': return True
	if data[4:7] == b'\x1a\x07\x01': return True
	return False
#end def findRAR(f, fpos)

def findLZMA(f, fpos):
	"""
	Given N bytes of data, determine whether it's a 7zip LZMA.
	7zip format is:
	signature '7', 'z', 0xBC, 0xAF, 0x27, 0x1C
	'7z\xbc\xaf\x27\x1c'
	"""
	data = f[fpos:fpos+6]
	if data == None or len(data) < 6:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:6] != b'7z\xbc\xaf\x27\x1c': return False
	return True
#end def findLZMA(f, fpos)

def findELF(f, fpos):
	"""
	Given N bytes of data, determine whether it's a ELF.
	ELF format is:
	magic \x7fELF
	The first 16 bytes are shared by 99.8% of the ELFs so far.
	7f 45 4c 46 02 01 01 00  00 00 00 00 00 00 00 00
	The only byte that is different is 5
	7f 45 4c 46 01 01 01 00  00 00 00 00 00 00 00 00
	"""
	data = f[fpos:fpos+4]
	if data == None or len(data) < 4:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	if data[:4] == b'\x7fELF':
		#if data[4] == '\x01': print '32-bit elf mayhaps?'
		return True
	return False
#end def findELF(f, fpos)

def findSH(f, fpos):
	"""
	Given N bytes of data, determine whether it's a SH.
	SH format is:
	magic #!/bin/sh
	"""
	data = f[fpos:fpos+11]
	if data == None or len(data) < 4:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	# This covers most right?
	if data[:4] != b'#!/b' and data[:4] != b'#!/u': return False
	if len(data) < 9: return POSSIBLE
	if data[:9] == b'#!/bin/sh': return True
	if data[:11] == b'#!/bin/bash': return True
	return POSSIBLE
#end def findSH(f, fpos)

def findPython(f, fpos):
	"""
	Given N bytes of data, determine whether it's a Python script.
	Script format is:
	magic #!/usr/bin/env python
	"""
	data = f[fpos:fpos+21]
	if data == None or len(data) < 4:
		#print 'Error: somehow data is not long enough:', len(data)
		return False
	#end if
	# This covers most right?
	if data[:4] != b'#!/u': return False
	if len(data) < 21: return POSSIBLE
	if data[:21] == b'#!/usr/bin/env python': return True
	return POSSIBLE
#end def findPython(f, fpos)

def findPDF(f, fpos):
	"""
	Given N bytes of data, determine whether it's a PDF.
	PDF format is very flexible, but it usually starts with the version in a 
	comment:
	%PDF-1.4
	"""
	data = f[fpos:fpos+12]
	if data == None or len(data) < 4:
		return False
	#end if
	if data[:4] != b'%PDF': return False
	return POSSIBLE
#end def findPDF(f, fpos)

def findBitTorrent(f, fpos):
	"""
	Given N bytes of data, determine whether it's a BitTorrent.
	BitTorrent format is fairly flexible, but it usually with d8:announce
	"""
	data = f[fpos:fpos+11]
	if data == None or len(data) < 11:
		return False
	#end if
	if data[:11] == b'd8:announce': return True
	return False
#end def findBitTorrent(f, fpos)

def findNullsoft(f, fpos):
	"""
	Given N bytes of data, determine whether it's a Nullsoft.
	Nullsoft can be extracted with 7z from p7zip.
	For example:
	7z l ~/Downloads/emerald-city-confidential_s1_l1_gF2911T1L1_d1467275361.exe
	"""
	data = f[fpos:fpos+12]
	if data == None or len(data) < 12:
		return False
	#end if
	if data[:12] == b'NullsoftInst': return True
	return False
#end def findNullsoft(f, fpos)

def findBOM(f, fpos):
	"""
	Given N bytes of data, determine whether it's a Unicode BOM.
	BOM stands for Byte order marker.
	"""
	data = f[fpos:fpos+3]
	if data == None or len(data) < 3: return False
	if data[:3] != b'\xef\xbb\xbf': return False
	return True
#end def findBOM(f, fpos)

def findHwra(f, fpos):
	"""
	Given N bytes of data, determine whether it's a Hammerwatch assets.bin. These files can be extracted with hwra1.py
	TODO: create parser and outputter from hwra1.py.
	"""
	data = f[fpos:fpos+4]
	if data == None or len(data) < 4: return False
	if data[:4] != b'HWRA': return False
	return True
#end def findHwra(f, fpos)

finders = {'gzip':findGzip, 'zlib':findZlib, 'png':findPNG, 'swf':findSWF, 
'bzip2':findBZip2, 'xz':findXZ, 'gif':findGIF, 'zip':findZIP, 'jpeg':findJPEG, 
'rar':findRAR, 'elf':findELF, 'sh':findSH, 'tar':findTAR, 'mp3':findMP3, 
'ogg':findOgg, 'wav':findWAV, 'avi':findAVI, 'py':findPython, 'lzma':findLZMA, 
'pdf':findPDF, 'torrent':findBitTorrent, 'flac':findFLAC, 'unicode':findBOM, 
'nullsoft':findNullsoft, 'hwra':findHwra, 'bmp':findBMP, 'tga':findTGA}

parsers = {'png':parsePNG, 'jpeg':parseJPEG, 'wav':parseWAV, 'ogg':parseOGG, 
'gzip':parseGzip, 'gif':parseGIF}

outputters = {'png':outputPNG, 'jpeg':outputJPEG, 'wav':outputWAV, 
'ogg':outputOGG, 'zlib':outputZlib, 'bmp':outputBMP}

"""parsers_todo = {'gzip':parseGzip, 'zlib':parseZlib, 'png':parsePNG, 'swf':parseSWF, 
'bzip2':parseBZip2, 'xz':parseXZ, 'gif':parseGIF, 'zip':parseZIP, 'jpeg':parseJPEG, 
'rar':parseRAR, 'elf':parseELF, 'sh':parseSH, 'tar':parseTAR, 'mp3':parseMP3, 
'ogg':parseOgg}"""

def getResult(filename, bestMatch, partialMatch, returnNone = False):
	prefix = filename + ': '
	if not filename: prefix = ''
	offset_str = ''
	if bestMatch:
		offset = bestMatch[1]
		if offset > 0: offset_str = ' (offset: %i)' % offset
		return prefix + bestMatch[0] + offset_str
	elif partialMatch:
		offset = partialMatch[1]
		if offset > 0: offset_str = ' (offset: %i)' % offset
		return prefix + partialMatch[0] + ' (possibly)' + offset_str
	#end if
	# It's easier to check for none than a string
	if returnNone: return None
	return prefix + 'unknown'
#end def getResult(filename, bestMatch, partialMatch)

def findx(f, filename='', seek_offset=0, search_len=1048576, multi=True, outputDirectory=None, printOut=False, verbose=False, outputZlib=False):
	"""
	f is either a mmap or a byte array.
	"""
	exitCode = 1
	bestMatch = None
	partialMatch = None
	results = []
	if type(f) == mmap.mmap:
		# Find how large the file is
		f.seek(0, 2)
		file_size = f.tell()
		if printOut: print('File size:', file_size)
		f.seek(0, 0)
	elif (sys.hexversion >= 0x3000000) and type(f) == str:
		raise ValueError("No strings allowed, convert to bytes please.")
	else:
		file_size = len(f)
	#end if
	y = 1
	# Look through N bytes
	for i in range(search_len):
		newMatch = False
		newBest = False
		newPartial = False
		# TODO: I forget why I put "not multi and" here.
		if i + 4 > file_size: break
		for pname, finder in list(finders.items()):
			#print pname, offset, 'currData[offset:]', repr(currData[offset:offset+4])
			found = finder(f, seek_offset + i)
			#print found
			if found == True:
				#print 'fo,o', file_offset, offset
				bestMatch = (pname, seek_offset + i, i)
				newMatch = newBest = True
				if not multi: break
			elif found == POSSIBLE or (verbose and found == UNLIKELY):
				partialMatch = (pname, seek_offset + i, i)
				newMatch = newPartial = True
			#end if
		#next pname, finder
		if newMatch:
			if verbose:
				# TODO: Get more data.
				if newBest and bestMatch[0] in parsers:
					if printOut: print(parsers[bestMatch[0]](f[bestMatch[1]:], 0))
				elif newPartial and partialMatch[0] in parsers:
					if printOut: print(parsers[partialMatch[0]](f[partialMatch[1]:], 0))
				#end if
			#end if
			if outputDirectory:
				#print '**o', newBest, bestMatch, newPartial, partialMatch
				if newBest and bestMatch[0] in outputters:
					if outputZlib or bestMatch[0] != 'zlib':
						x = outputters[bestMatch[0]](f, bestMatch[1])
						open((outputDirectory + '/file_%i.' % (y)) + bestMatch[0], 'wb').write(x)
						x = None
						y += 1
					#end if
				elif newPartial and partialMatch[0] in outputters:
					if outputZlib or partialMatch[0] != 'zlib':
						x = outputters[partialMatch[0]](f, partialMatch[1])
						open((outputDirectory + '/file_%i.' % y) + partialMatch[0], 'wb').write(x)
						x = None
						y += 1
					#end if
				#end if
			#end if
			if multi:
				if printOut:
					r = getResult(filename, newBest and bestMatch, newPartial and partialMatch, True)
					if r:
						print(r)
					#end if
				else:
					if newBest:
						results.append(bestMatch)
					else:
						results.append(partialMatch)
					#end if
				#end if
				exitCode = 0
			else:
				if bestMatch:
					break
				else:
					# Print possible matches.
					print(getResult(filename, bestMatch, partialMatch))
				#end if
			#end if
		#end if
	#next i
	if not multi:
		if printOut:
			print(getResult(filename, bestMatch, partialMatch))
		else:
			return bestMatch or partialMatch
		#end if
		exitCode = 0
	#end if
	if printOut:
		return exitCode
	#end if
	return results
#end def findx(f, filename='', seek_offset=0, search_len=1048576, multi=True, outputDirectory=None, printOut=False, verbose=False, outputZlib=False)

def main():
	import os
	import stat
	if len(sys.argv) < 2:
		usage()
		exit(1)
	#end if
	outputDirectory = None
	outputZlib = False
	multi = False
	verbose = False
	seek_offset = 0
	search_len = 1024*1024
	exitCode = 1
	# TODO: use getopt instead, this is getting silly
	if '-o' in sys.argv:
		opos = sys.argv.index('-o') + 1
		if opos >= len(sys.argv):
			usage()
			exit(1)
		#end if
		seek_offset = int(sys.argv[opos])
		opos -= 1
		sys.argv.pop(opos)
		sys.argv.pop(opos)
	#end if
	if '-d' in sys.argv:
		opos = sys.argv.index('-d') + 1
		if opos >= len(sys.argv):
			usage()
			exit(1)
		#end if
		outputDirectory = sys.argv[opos]
		opos -= 1
		sys.argv.pop(opos)
		sys.argv.pop(opos)
		if not os.path.exists(outputDirectory):
			os.mkdir(outputDirectory)
		#end if
		file_stat = os.stat(outputDirectory)
		if not stat.S_ISDIR(file_stat.st_mode):
			print(outputDirectory + ': ' + 'not directory')
			exit(1)
		#end if
	#end if
	if '-z' in sys.argv:
		outputZlib = True
		sys.argv.remove('-z')
	#end if
	if '-m' in sys.argv:
		multi = True
		sys.argv.remove('-m')
	#end if
	if '-v' in sys.argv:
		verbose = True
		sys.argv.remove('-v')
	#end if
	if '-l' in sys.argv:
		lpos = sys.argv.index('-l') + 1
		if lpos >= len(sys.argv):
			usage()
			exit(1)
		#end if
		search_len = int(sys.argv[lpos])
		lpos -= 1
		sys.argv.pop(lpos)
		sys.argv.pop(lpos)
	#end if
	filenames = sys.argv[1:]
	for filename in filenames:
		file_offset = seek_offset
		offset = 0
		if not os.path.exists(filename):
			print(filename + ': ' + 'does not exist')
			continue
		#end if
		file_stat = os.stat(filename)
		if stat.S_ISDIR(file_stat.st_mode):
			print(filename + ': ' + 'directory')
			continue
		#end if
		try:
			fobj = open(filename, 'rb')
		except IOError as e:
			print(filename + ': ' + str(e))
			continue
		#end if
		f = mmap.mmap(fobj.fileno(), 0, prot=mmap.PROT_READ)
		exitCode_i = findx(f, filename, seek_offset, search_len, multi, outputDirectory, True, verbose, outputZlib)
		# Only set exitCode to 0, not away.
		if exitCode_i == 0:
			exitCode = 0
		#end if
	#next filename
	exit(exitCode)
#end def main()

if __name__ == '__main__':
	main()
#end if
