#!/usr/bin/env python
"""
Histogram of bytes in multiple files
by Javantea
Mar 24, 2012
Upgraded to Python3 Feb 19, 2015

Produces a data file ready for reverse engineering a proprietary file format.
See hist_xz1.txt for an example of reverse enginering xz.
"""
from __future__ import print_function
import sys
import histogram1
import os
import stat

def usage():
	print("Usage: hist_files.py [-o offset] file.ext [...] > file.dat")
#end def usage()

def main():
	if len(sys.argv) < 2:
		usage()
		exit(1)
	#end if
	offset = 0
	if '-o' in sys.argv:
		opos = sys.argv.index('-o') + 1
		if opos >= len(sys.argv):
			usage()
			exit(1)
		#end if
		offset = int(sys.argv[opos])
		opos -= 1
		sys.argv.pop(opos)
		sys.argv.pop(opos)
	#end if
	filenames = sys.argv[1:]
	b = [b'' for i in range(16)]
	for filename in filenames:
		if not os.path.exists(filename):
			print(filename + ': ' + 'does not exist')
			continue
		#end if
		file_stat = os.stat(filename)
		if stat.S_ISDIR(file_stat.st_mode):
			print(filename + ': ' + 'directory')
			continue
		#end if
		try:
			f = open(filename, 'rb')
		except IOError as e:
			print(filename + ': ' + str(e))
			continue
		#end if
		if offset > 0:
			f.seek(offset, 0)
		#end if
		a = f.read(16)
		for i in range(len(a)):
			b[i] += bytes([a[i]])
		#next i
	#next filename
	#print b
	# Likeliness of X being in the 1 spot:
	hists = [histogram1.bufferBinaryHistogram(b[i]) for i in range(16)]
	#print hists
	for j, hist in enumerate(hists):
		print('# byte ', j)
		for i, f in enumerate(hist):
			if f > 0: print(i, f, repr(bytes([i])))
		#next f
	#next hist
#end def main()

if __name__ == '__main__':
	main()
#end if
